/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>
#include <string.h>

#include "glass.h"

/* Make a new active point */
GlassActivePoint *_glassActivePointNew(void)
{
  GlassActivePoint *apoint = malloc(sizeof(GlassActivePoint));

  if(apoint == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for active point\n");
    return NULL;
  }

  apoint->name = NULL;
  _glassActivePointSetName(apoint, "");

  /* Default position on the origion */
  apoint->pos[0] = 0.0;
  apoint->pos[1] = 0.0;
  apoint->pos[2] = 0.0;

  /* Default point down the x-axis */
  apoint->initial_dir[0] = 1.0;
  apoint->initial_dir[1] = 0.0;
  apoint->initial_dir[2] = 0.0;
  apoint->dir[0] = 1.0;
  apoint->dir[1] = 0.0;
  apoint->dir[2] = 0.0;

  apoint->next = NULL;

  return apoint;
}

/* Set the active point's name */
void _glassActivePointSetName(GlassActivePoint *apoint, const char *name)
{
  free(apoint->name);
  if(name == NULL)
    apoint->name = strdup("");
  else
    apoint->name = strdup(name);
}

/* Get the active point's name */
const char *_glassActivePointGetName(GlassActivePoint *apoint)
{
  return apoint->name;
}

/* Set the initial direction of the active point */
void _glassActivePointSetInitDir(GlassActivePoint *apoint, GLfloat dx, GLfloat dy, GLfloat dz)
{
  apoint->initial_dir[0] = dx;
  apoint->initial_dir[1] = dy;
  apoint->initial_dir[2] = dz;
}

/* Set the initial direction of the active point with an array (vector) */
void _glassActivePointSetInitDirv(GlassActivePoint *apoint, GLfloat *dir)
{
  apoint->initial_dir[0] = dir[0];
  apoint->initial_dir[1] = dir[1];
  apoint->initial_dir[2] = dir[2];
}

/* Get the initial direction of the active point */
GLfloat *_glassActivePointGetInitDir(GlassActivePoint *apoint)
{
  return apoint->initial_dir;
}

/* Get the current position  of the active point */
GLfloat *_glassActivePointGetPos(GlassActivePoint *apoint)
{
  return apoint->pos;
}

/* Get the current direction of the active point */
GLfloat *_glassActivePointGetDir(GlassActivePoint *apoint)
{
  return apoint->dir;
}

/* Update the position, and direction of the active point by using the current
   openGL modelview matrix */
void _glassActivePointUpdate(GlassActivePoint *apoint)
{
  GLfloat m[16], *pos = apoint->pos, *idir = apoint->initial_dir, *dir = apoint->dir;

  /* Get the modelview matrix from openGL */
  glGetFloatv(GL_MODELVIEW_MATRIX, m);

  /* Multiply the matrix by the vector [0 0 0 1] to find where the origion
     translates to -- the absolute position of the point */
  pos[0] = m[12];
  pos[1] = m[13];
  pos[2] = m[14];

  /* Multiply the matrix by the vector [idir[0] idir[1] idir[2] 1] and subtract
     the above position to find where a direction vector points to -- the new
     direction vector for the point */
  dir[0] = idir[0]*m[0] + idir[1]*m[4] + idir[2]*m[8];
  dir[1] = idir[0]*m[1] + idir[1]*m[5] + idir[2]*m[9];
  dir[2] = idir[0]*m[2] + idir[1]*m[6] + idir[2]*m[10];
}

/* Free the active point */
void _glassFreeActivePoint(GlassActivePoint *apoint)
{
  free(apoint->name);
  free(apoint);
}
