/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* This file contains the functions used externally to access the library */

#include "glass.h"

/* Return a string for this version of glass */
char *glassGetVersion(void)
{
  static char version[] = VERSION_STRING;

  return version;
}

/* Load an object */
GlassObject *glassLoadObject(const char *fname)
{
  return _glassLoadObject(NULL, fname);
}

/* Make a new library */
GlassLibrary *glassLibraryNew(const char *name)
{
  return _glassLibraryNew(name);
}

/* Load an object into a library */
GlassObject *glassLibraryLoadObject(GlassLibrary *library, const char *fname)
{
  return _glassLoadObject(library, fname);
}

/* Return the index of a variable with the name `name', return -1 if not found */
int glassFindVariable(GlassObject *object, const char *name)
{
  GlassVariable *variable = _glassObjectFindVariable(object, name);
  int i;

  if(variable == NULL)
    return -1;

  /* Find the index */
  for(i = 0; i < object->nvariables && object->variable_array[i] != variable; i++);

  return i;
}

/* Get the current value of a variable */
GLfloat glassGetVariableValue(GlassObject *object, int variable)
{
  if(variable >= object->nvariables || variable < 0)
    return FALSE;

  return _glassVariableGetValue(object->variable_array[variable]);
}

/* Get a pointer to the current value of a variable */
GLfloat *glassGetVariableValuev(GlassObject *object, int variable)
{
  if(variable >= object->nvariables || variable < 0)
    return FALSE;

  return _glassVariableGetValuev(object->variable_array[variable]);
}

/* Get the minimum value a variable can take */
GLfloat glassGetVariableMin(GlassObject *object, int variable)
{
  if(variable >= object->nvariables || variable < 0)
    return FALSE;

  return _glassVariableGetMin(object->variable_array[variable]);
}

/* Get the maximum value a variable can take */
GLfloat glassGetVariableMax(GlassObject *object, int variable)
{
  if(variable >= object->nvariables || variable < 0)
    return FALSE;

  return _glassVariableGetMax(object->variable_array[variable]);
}

/* Set a variable, return FALSE if the variable doesn't exist */
int glassSetVariable(GlassObject *object, int variable, GLfloat value)
{
  if(variable >= object->nvariables || variable < 0)
    return FALSE;

  _glassVariableSetValue(object->variable_array[variable], value);
  object->variables_changed = TRUE;

  return TRUE;
}

/* Increment a variable, return FALSE if the variable doesn't exist */
int glassIncVariable(GlassObject *object, int variable, GLfloat dvalue)
{
  if(variable >= object->nvariables || variable < 0)
    return FALSE;

  _glassVariableIncValue(object->variable_array[variable], dvalue);
  object->variables_changed = TRUE;

  return TRUE;
}

/* Find an active point */
int glassFindActivePoint(GlassObject *object, const char *name)
{
  GlassActivePoint *apoint = _glassObjectFindActivePoint(object, name);
  int i;

  if(apoint == NULL)
    return -1;

  /* Find the index */
  for(i = 0; i < object->npoints && object->apoint_array[i] != apoint; i++);

  return i;
}

/* Get the position of an active point */
GLfloat *glassGetActivePointPos(GlassObject *object, int apoint)
{
  if(apoint >= object->npoints || apoint < 0)
    return NULL;

  /* Update the active points if necessary */
  if(object->variables_changed)
  {
    _glassObjectUpdateActivePoints(object);
    object->variables_changed = FALSE;
  }

  return _glassActivePointGetPos(object->apoint_array[apoint]);
}

/* Get the direction of an active point */
GLfloat *glassGetActivePointDir(GlassObject *object, int apoint)
{
  if(apoint >= object->npoints || apoint < 0)
    return NULL;

  /* Update the active points if necessary */
  if(object->variables_changed)
  {
    _glassObjectUpdateActivePoints(object);
    object->variables_changed = FALSE;
  }

  return _glassActivePointGetDir(object->apoint_array[apoint]);
}

/* Draw the object, return FALSE if error occurs */
int glassDrawObject(GlassObject *object)
{
   /* Update the active points if necessary */
  if(object->variables_changed)
  {
    _glassObjectUpdateActivePoints(object);
    object->variables_changed = FALSE;
  }
   
  return _glassDrawObject(object);
}

/* Return the index of a component with the name `name', return -1 if not found */
int glassFindComponent(GlassObject *object, const char *name)
{
  GlassTriangleList *tlist = _glassObjectFindList(object, name);
  int i;

  if(tlist == NULL)
    return -1;

  /* Find the index */
  for(i = 0; i < object->nlists && object->list_array[i] != tlist; i++);

  return i;
}

/* Toggle whether or not this component (triangle list), is drawn or not,
   visible = TRUE = draw component, visible = FALSE = hide component.
   Return FALSE if component doesn't exist */
int glassSetComponentVisibility(GlassObject *object, int component, int visible)
{
  if(component >= object->nlists || component < 0)
    return FALSE;

  _glassTriangleListSetVisibility(object->list_array[component], visible);
  return TRUE;
}

/* Draw one component (triangle list) of the object, return FALSE if component doesn't exist */
int glassDrawComponent(GlassObject *object, int component)
{
  if(component >= object->nlists || component < 0)
    return FALSE;

  if(!_glassDrawTriangleList(object->list_array[component]))
    return FALSE;
  return _glassDrawAlphaTriangleList(object->list_array[component]);
}

/* Remove an object and all its components etc */
void glassFreeObject(GlassObject *object)
{
  GlassTexture *tex_ptr = object->textures;
  GlassMaterial *mat_ptr = object->materials;
  GlassTriangleList *list_ptr = object->lists;
  GlassVariable *var_ptr = object->variables;
  GlassActivePoint *apoint_ptr = object->points;
  void *tmp_ptr;

  if(object == NULL)
    return;

  /* Free the textures */
  while(tex_ptr != NULL)
  {
    tmp_ptr = (void *) (tex_ptr->next);
    _glassFreeTexture(tex_ptr);
    tex_ptr = (GlassTexture *) tmp_ptr;
  }

  /* Free the lists */
  while(mat_ptr)
  {
    tmp_ptr = (void *) (mat_ptr->next);
    _glassFreeMaterial(mat_ptr);
    mat_ptr = (GlassMaterial *) tmp_ptr;
  }

  /* Free the lists */
  while(list_ptr)
  {
    tmp_ptr = (void *) (list_ptr->next);
    _glassFreeTriangleListChild(list_ptr);
    list_ptr = (GlassTriangleList *) tmp_ptr;
  }

  /* Free the variables */
  while(var_ptr)
  {
    tmp_ptr = (void *) (var_ptr->next);
    _glassFreeVariable(var_ptr);
    var_ptr = (GlassVariable *) tmp_ptr;
  }

  /* Free the active points */
  while(apoint_ptr)
  {
    tmp_ptr = (void *) (apoint_ptr->next);
    _glassFreeActivePoint(apoint_ptr);
    apoint_ptr = (GlassActivePoint *) tmp_ptr;
  }

  /* Free the components */
  _glassFreeComponentr(object->child);

  /* Finally, free the object */
  _glassFreeObject(object);
}

/* Remove a library and all its objects */
void glassFreeLibrary(GlassLibrary *library)
{
  GlassLibraryEntry *entry;

  if(!library)
    return;

  entry = library->first;
  while(entry)
  {
    glassFreeObject(entry->object);
    entry->object = NULL;
    entry = entry->next;
  }

  _glassFreeLibrary(library);
}
