/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>
#include <string.h>

#include "glass.h"

/* Return an empty library */
GlassLibrary *_glassLibraryNew(const char *name)
{
  GlassLibrary *library;

  library = malloc(sizeof(GlassLibrary));
  if(!library)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for GLASS library (%s:%d)\n",
	    __FILE__, __LINE__);
    return NULL;
  }

  library->name = NULL;
  _glassLibrarySetName(library, name);

  library->first = NULL;

  return library;
}

/* Set the name of the library */
int _glassLibrarySetName(GlassLibrary *library, const char *name)
{
  if(!library)
    return FALSE;

  if(library->name)
    free(library->name);

  if(name)
    library->name = strdup(name);
  else
    library->name = strdup("");

  return TRUE;
}

/* Return the name of the library */
char *_glassLibraryGetName(GlassLibrary *library)
{
  if(!library)
    return NULL;

  return library->name;
}

/* Add a GLASS object to a GLASS library */
int _glassLibraryAddObject(GlassLibrary *library, GlassObject *object)
{
  GlassLibraryEntry *entry;

  if(!library || !object)
    return FALSE;

  /* Make a new entry */
  entry = malloc(sizeof(GlassLibraryEntry));
  if(!entry)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for GLASS library entry (%s:%d)\n",
	    __FILE__, __LINE__);
    return FALSE;
  }

  entry->object = object;

  /* Add to front of the list */
  entry->last = NULL;
  entry->next = library->first;
  if(entry->next)
    entry->next->last = entry;
  library->first = entry;

  object->library = library;

  return TRUE;
}

/* Find a GLASS object in a GLASS library, return NULL if not found */
GlassObject *_glassLibraryFindObject(GlassLibrary *library, const char *name)
{
  GlassLibraryEntry *entry_ptr;

  if(!library || !name)
    return NULL;

  entry_ptr = library->first;
  while(entry_ptr)
  {
    if(entry_ptr->object)
    {
      if(!strcmp(name, _glassObjectGetName(entry_ptr->object)))
	return entry_ptr->object;
    }

    entry_ptr = entry_ptr->next;
  }

  /* Not found */
  return NULL;
}

/* Find a texture from any of the objects in the library */
GlassTexture *_glassLibraryFindTexture(GlassLibrary *library, const char *name)
{
  GlassLibraryEntry *entry_ptr;
  GlassTexture *texture = NULL;

  if(!library || !name)
    return NULL;

  entry_ptr = library->first;
  while(entry_ptr && !texture)
  {
    if(entry_ptr->object)
    {
      /* Disable the library to avoid recursion */
      entry_ptr->object->library = NULL;
      texture = _glassObjectFindTexture(entry_ptr->object, name);
      entry_ptr->object->library = library;
    }

    entry_ptr = entry_ptr->next;
  }

  return texture;
}

/* Find a material from any of the objects in the library */
GlassMaterial *_glassLibraryFindMaterial(GlassLibrary *library, const char *name)
{
  GlassLibraryEntry *entry_ptr;
  GlassMaterial *material = NULL;

  if(!library || !name)
    return NULL;

  entry_ptr = library->first;
  while(entry_ptr && !material)
  {
    if(entry_ptr->object)
    {
      /* Disable the library to avoid recursion */
      entry_ptr->object->library = NULL;
      material = _glassObjectFindMaterial(entry_ptr->object, name);
      entry_ptr->object->library = library;
    }

    entry_ptr = entry_ptr->next;
  }

  return material;
}

/* Find a triangle list from any of the objects in the library */
GlassTriangleList *_glassLibraryFindList(GlassLibrary *library, const char *name)
{
  GlassLibraryEntry *entry_ptr;
  GlassTriangleList *tlist = NULL;

  if(!library || !name)
    return NULL;

  entry_ptr = library->first;
  while(entry_ptr && !tlist)
  {
    if(entry_ptr->object)
    {
      /* Disable the library to avoid recursion */
      entry_ptr->object->library = NULL;
      tlist = _glassObjectFindList(entry_ptr->object, name);
      entry_ptr->object->library = library;
    }

    entry_ptr = entry_ptr->next;
  }

  return tlist;
}

/* Find a variable from any of the objects in the library */
GlassVariable *_glassLibraryFindVariable(GlassLibrary *library, const char *name)
{
  GlassLibraryEntry *entry_ptr;
  GlassVariable *variable = NULL;

  if(!library || !name)
    return NULL;

  entry_ptr = library->first;
  while(entry_ptr && !variable)
  {
    if(entry_ptr->object)
    {
      /* Disable the library to avoid recursion */
      entry_ptr->object->library = NULL;
      variable = _glassObjectFindVariable(entry_ptr->object, name);
      entry_ptr->object->library = library;
    }

    entry_ptr = entry_ptr->next;
  }

  return variable;
}
 
/* Find a active point from any of the objects in the library */
GlassActivePoint *_glassLibraryFindActivePoint(GlassLibrary *library, const char *name)
{
  GlassLibraryEntry *entry_ptr;
  GlassActivePoint *apoint = NULL;

  if(!library || !name)
    return NULL;

  entry_ptr = library->first;
  while(entry_ptr && !apoint)
  {
    if(entry_ptr->object)
    {
      /* Disable the library to avoid recursion */
      entry_ptr->object->library = NULL;
      apoint = _glassObjectFindActivePoint(entry_ptr->object, name);
      entry_ptr->object->library = library;
    }

    entry_ptr = entry_ptr->next;
  }

  return apoint;
}

/* Remove a GLASS object from a GLASS library */
int _glassLibraryRemoveObject(GlassLibrary *library, GlassObject *object)
{
  GlassLibraryEntry *entry_ptr;

  if(!library || !object)
    return FALSE;

  /* Find the entry with this object */
  entry_ptr = library->first;
  while(entry_ptr && entry_ptr->object != object)
    entry_ptr = entry_ptr->next;

  /* Not found */
  if(!entry_ptr)
    return FALSE;

  /* Delink */
  if(entry_ptr->next)
    entry_ptr->next->last = entry_ptr->last;
  if(entry_ptr->last)
    entry_ptr->last->next = entry_ptr->next;
  else
    library->first = entry_ptr->next;

  /* Isolate */
  entry_ptr->next = NULL;
  entry_ptr->last = NULL;
  object->library = NULL;

  /* TODO: Remove links from other objects to this object */

  return TRUE;
}

/* Free a library entry */
int _glassFreeLibraryEntry(GlassLibraryEntry *entry)
{
  if(!entry)
    return FALSE;

  free(entry);

  return TRUE;
}

/* Free a library */
int _glassFreeLibrary(GlassLibrary *library)
{
  GlassLibraryEntry *entry, *tmp;

  if(!library)
    return FALSE;

  entry = library->first;
  while(entry)
  {
    tmp = entry->next;

    if(entry->object)
      entry->object->library = NULL;
    _glassFreeLibraryEntry(entry);

    entry = tmp;
  }

  if(library->name)
    free(library->name);
  free(library);

  return TRUE;
}
