/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>
#include <string.h>

#include "glass.h"

/* Returns an empty object */
GlassObject *_glassObjectNew(const char *name)
{
  GlassObject *object;

  object = malloc(sizeof(GlassObject));
  if(!object)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for object\n");
    return NULL;
  }

  object->name = NULL;
  _glassObjectSetName(object, name);
  object->library = NULL;

  object->nlists = 0;
  object->nvariables = 0;
  object->npoints = 0;
  object->list_array = NULL;
  object->variable_array = NULL;
  object->apoint_array = NULL;

  object->variables_changed = TRUE;

  object->textures = NULL;
  object->materials = NULL;
  object->lists = NULL;
  object->variables = NULL;
  object->points = NULL;
  object->child = NULL;

  return object;
}

/* Set the object's name */
void _glassObjectSetName(GlassObject *object, const char *name)
{
  if(!object)
    return;

  if(object->name)
    free(object->name);

  if(name)
    object->name = strdup(name);
  else
    object->name = strdup("");
}

/* Get the object's name */
const char *_glassObjectGetName(GlassObject *object)
{
  if(!object)
    return NULL;

  return object->name;
}

/* Add a texture into the object */
void _glassObjectAddTexture(GlassObject *object, GlassTexture *texture)
{
  GlassTexture **texture_ptr;

  if(!object || !texture)
    return;

  /* Find the end of the list */
  texture_ptr = &object->textures;
  while(*texture_ptr)
    texture_ptr = &(*texture_ptr)->next;

  *texture_ptr = texture;
}

/* Find a texture with the name `name', return NULL if not found */
GlassTexture *_glassObjectFindTexture(GlassObject *object, const char *name)
{
  GlassTexture *texture_ptr;

  if(!object || !name)
    return NULL;

  texture_ptr = object->textures;
  while(texture_ptr)
  {
    if(!strcmp(_glassTextureGetName(texture_ptr), name))
      return texture_ptr;

    texture_ptr = texture_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(object->library)
    return _glassLibraryFindTexture(object->library, name);

  return NULL;
}

/* Remove a texture from a object, returning FALSE if the texture isn't in the object */
int _glassObjectRemoveTexture(GlassObject *object, GlassTexture *texture)
{
  GlassTexture *texture_ptr;

  if(!object || !object->textures || !texture)
    return FALSE;

  /* If is the first entry */
  texture_ptr = object->textures;
  if(texture_ptr == texture)
    object->textures = texture->next;
  else {
    /* Find the texture before this one */
    while(texture_ptr && texture_ptr->next != texture)
      texture_ptr = texture_ptr->next;

    /* If didn't find it */
    if(!texture_ptr)
      return FALSE;

    /* Skip out the list */
    texture_ptr->next = texture->next;
  }

  /* Texture is now isolated */
  texture->next = NULL;

  return TRUE;
}

/* Add a material into the object */
void _glassObjectAddMaterial(GlassObject *object, GlassMaterial *material)
{
  GlassMaterial **material_ptr;

  if(!object || !material)
    return;

  /* Find the end of the list */
  material_ptr = &object->materials;
  while(*material_ptr)
    material_ptr = &(*material_ptr)->next;

  *material_ptr = material;
}

/* Find a material with the name `name', return NULL if not found */
GlassMaterial *_glassObjectFindMaterial(GlassObject *object, const char *name)
{
  GlassMaterial *material_ptr;

  if(!object || !name)
    return NULL;

  material_ptr = object->materials;
  while(material_ptr)
  {
    if(!strcmp(_glassMaterialGetName(material_ptr), name))
      return material_ptr;

    material_ptr = material_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(object->library)
    return _glassLibraryFindMaterial(object->library, name);

  return NULL;
}

/* Remove a material from a object, returning FALSE if the material isn't in the object */
int _glassObjectRemoveMaterial(GlassObject *object, GlassMaterial *material)
{
  GlassMaterial *material_ptr;

  if(!object || !object->materials || !material)
    return FALSE;

  /* If is the first entry */
  material_ptr = object->materials;
  if(material_ptr == material)
    object->materials = material->next;
  else {
    /* Find the material before this one */
    while(material_ptr && material_ptr->next != material)
      material_ptr = material_ptr->next;

    /* If didn't find it */
    if(!material_ptr)
      return FALSE;

    /* Skip out the list */
    material_ptr->next = material->next;
  }

  /* Material is now isolated */
  material->next = NULL;

  return TRUE;
}

/* Add a triangle list into the object */
void _glassObjectAddList(GlassObject *object, GlassTriangleList *tlist)
{
  GlassTriangleList **tlist_ptr;

  if(!object || !tlist)
    return;

  /* Find the end of the list */
  tlist_ptr = &object->lists;
  while(*tlist_ptr)
    tlist_ptr = &(*tlist_ptr)->next;

  *tlist_ptr = tlist;

  /* Update the shortcut array */
  _glassObjectBuildListArray(object);
}

/* Find a triangle list with the name `name', return NULL if not found */
GlassTriangleList *_glassObjectFindList(GlassObject *object, const char *name)
{
  GlassTriangleList *tlist_ptr;

  if(!object || !name)
    return NULL;

  tlist_ptr = object->lists;
  while(tlist_ptr)
  {
    if(!strcmp(_glassTriangleListGetName(tlist_ptr), name))
      return tlist_ptr;

    tlist_ptr = tlist_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(object->library)
    return _glassLibraryFindList(object->library, name);

  return NULL;
}

/* Remove a triangle list from a object, returning FALSE if the list isn't in the object */
int _glassObjectRemoveList(GlassObject *object, GlassTriangleList *tlist)
{
  GlassTriangleList *tlist_ptr;

  /* If no lists */
  if(!object || !object->lists || !tlist)
    return FALSE;

  /* If is the first entry */
  tlist_ptr = object->lists;
  if(tlist_ptr == tlist)
    object->lists = tlist->next;
  else
  {
    /* Find the list before this one */
    while(tlist_ptr && tlist_ptr->next != tlist)
      tlist_ptr = tlist_ptr->next;

    /* If didn't find it */
    if(!tlist_ptr)
      return FALSE;

    /* Skip out the list */
    tlist_ptr->next = tlist->next;
  }

  /* List is now isolated */
  tlist->next = NULL;

  /* Update the shortcut array */
  _glassObjectBuildListArray(object);

  return TRUE;
}

/* Add a variable into the object */
void _glassObjectAddVariable(GlassObject *object, GlassVariable *variable)
{
  GlassVariable **variable_ptr;

  if(!object || !variable)
    return;

  /* Find the end of the list */
  variable_ptr = &object->variables;
  while(*variable_ptr)
    variable_ptr = &(*variable_ptr)->next;

  *variable_ptr = variable;

  /* Update the shortcut array */
  _glassObjectBuildVariableArray(object);
}

/* Find a variable with the name `name', return NULL if not found */
GlassVariable *_glassObjectFindVariable(GlassObject *object, const char *name)
{
  GlassVariable *var_ptr;

  if(!object || !name)
    return NULL;

  var_ptr = object->variables;
  while(var_ptr)
  {
    if(!strcmp(_glassVariableGetName(var_ptr), name))
      return var_ptr;

    var_ptr = var_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(object->library)
    return _glassLibraryFindVariable(object->library, name);

  return NULL;
}

/* Remove a variable from a object, returning FALSE if the variable isn't in the object */
int _glassObjectRemoveVariable(GlassObject *object, GlassVariable *variable)
{
  GlassVariable *var_ptr;

  if(!object || !object->variables || !variable)
    return FALSE;

  /* If is the first entry */
  var_ptr = object->variables;
  if(var_ptr == variable)
    object->variables = variable->next;
  else {
    /* Find the list before this one */
    while(var_ptr && var_ptr->next != variable)
      var_ptr = var_ptr->next;

    /* If didn't find it */
    if(!var_ptr)
      return FALSE;    

    /* Skip out the list */
    var_ptr->next = variable->next;
  }

  /* Variable is now isolated */
  variable->next = NULL;

  /* Update the shortcut array */
  _glassObjectBuildVariableArray(object);

  return TRUE;
}

/* Add an active point into the object */
void _glassObjectAddActivePoint(GlassObject *object, GlassActivePoint *apoint)
{
  GlassActivePoint **apoint_ptr;

  if(!object || !apoint)
    return;

  /* Find the end of the list */
  apoint_ptr = &object->points;
  while(*apoint_ptr)
    apoint_ptr = &(*apoint_ptr)->next;

  *apoint_ptr = apoint;

  /* Update the shortcut array */
  _glassObjectBuildActivePointArray(object);
}

/* Find a active point with the name `name', return NULL if not found */
GlassActivePoint *_glassObjectFindActivePoint(GlassObject *object, const char *name)
{
  GlassActivePoint *apoint_ptr;

  if(!object || !name)
    return NULL;

  apoint_ptr = object->points;
  while(apoint_ptr)
  {
    if(!strcmp(_glassActivePointGetName(apoint_ptr), name))
      return apoint_ptr;

    apoint_ptr = apoint_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(object->library)
    return _glassLibraryFindActivePoint(object->library, name);

  return NULL;
}

/* Remove a active point from a object, returning FALSE if the active point isn't
   in the object */
int _glassObjectRemoveActivePoint(GlassObject *object, GlassActivePoint *apoint)
{
  GlassActivePoint *apoint_ptr;

  if(!object || !object->points || !apoint)
    return FALSE;

  /* If is the first entry */
  apoint_ptr = object->points;
  if(apoint_ptr == apoint)
    object->points = apoint->next;
  else
  {
    /* Find the apoint before this one */
    while(apoint_ptr && apoint_ptr->next != apoint)
      apoint_ptr = apoint_ptr->next;

    /* If didn't find it */
    if(!apoint_ptr)
      return FALSE;

    /* Skip out the list */
    apoint_ptr->next = apoint->next;
  }

  /* Active point is now isolated */
  apoint->next = NULL;

  /* Update the shortcut array */
  _glassObjectBuildActivePointArray(object);

  return TRUE;
}

/* Add a child component to the object */
void _glassObjectAddChild(GlassObject *object, GlassComponent *component)
{
  if(!object || !component)
    return;

  /* If already has a child */
  if(object->child)
    _glassComponentAddSibling(object->child, component);
  else
    object->child = component;
}

/* Remove a child from an object, returning FALSE if child not found */
int _glassObjectRemoveChild(GlassObject *object, GlassComponent *child)
{
  if(!object || !child)
    return FALSE;

  if(object->child == child)
    object->child = child->next;
  else
    return _glassComponentRemoveSibling(object->child, child);

  return TRUE;
}

/* (Re)Build both the variable and triangle list arrays */
void _glassObjectBuildArrays(GlassObject *object)
{
  if(!object)
    return;

  _glassObjectBuildListArray(object);
  _glassObjectBuildVariableArray(object);
  _glassObjectBuildActivePointArray(object);
}

/* Build an array of pointers to variables, for quick access */
void _glassObjectBuildVariableArray(GlassObject *object)
{
  GlassVariable *var_ptr;
  int i, nvariables = 0;

  if(!object)
    return;

  /* Count the number of variables */
  var_ptr = object->variables;
  while(var_ptr)
  {
    nvariables++;
    var_ptr = var_ptr->next;
  }

  /* Resize the array */
  if(nvariables == 0)
  {
    free(object->variable_array);
    object->variable_array = NULL;
  }
  else
    object->variable_array = realloc(object->variable_array,
				     nvariables * sizeof(GlassVariable *));

  /* If no variables, set array to NULL */
  if(!nvariables)
    object->variable_array = NULL;

  /* Store pointers to the variables in order */
  var_ptr = object->variables;
  for(i = 0; i < nvariables; i++)
  {
    object->variable_array[i] = var_ptr;
    var_ptr = var_ptr->next;
  }

  /* Remeber how many variables there are */
  object->nvariables = nvariables;
}

/* Build an array of pointers to triangle lists, for quick access */
void _glassObjectBuildListArray(GlassObject *object)
{
  GlassTriangleList *tlist_ptr;
  int i, nlists = 0;

  if(!object)
    return;

  /* Count the number of lists */
  tlist_ptr = object->lists;
  while(tlist_ptr)
  {
    nlists++;
    tlist_ptr = tlist_ptr->next;
  }

  /* Resize the array */
  if(nlists == 0)
  {
    free(object->list_array);
    object->list_array = NULL;
  }
  else
    object->list_array = realloc(object->list_array, nlists * sizeof(GlassTriangleList *));

  /* If no lists, set array to NULL */
  if(!nlists)
    object->list_array = NULL;

  /* Store pointers to the lists in order */
  tlist_ptr = object->lists;
  for(i = 0; i < nlists; i++)
  {
    object->list_array[i] = tlist_ptr;
    tlist_ptr = tlist_ptr->next;
  }

  /* Remeber how many lists there are */
  object->nlists = nlists;
}

/* Build an array of pointers to active points, for quick access */
void _glassObjectBuildActivePointArray(GlassObject *object)
{
  GlassActivePoint *apoint_ptr;
  int i, npoints = 0;

  if(!object)
    return;

  /* Count the number of lists */
  apoint_ptr = object->points;
  while(apoint_ptr)
  {
    npoints++;
    apoint_ptr = apoint_ptr->next;
  }

  /* Resize the array */
  if(npoints == 0)
  {
    free(object->apoint_array);
    object->apoint_array = NULL;
  }
  else
    object->apoint_array = realloc(object->apoint_array, npoints * sizeof(GlassActivePoint *));

  /* If no lists, set array to NULL */
  if(!npoints)
    object->apoint_array = NULL;

  /* Store pointers to the lists in order */
  apoint_ptr = object->points;
  for(i = 0; i < npoints; i++)
  {
    object->apoint_array[i] = apoint_ptr;
    apoint_ptr = apoint_ptr->next;
  }

  /* Remeber how many lists there are */
  object->npoints = npoints;
}

/* Go through the object and update the active points by doing the transforms */
void _glassObjectUpdateActivePoints(GlassObject *object)
{
  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();

  glLoadIdentity();

  _glassComponentUpdateActivePointsr(object->child);

  glPopMatrix();
}

/* Convert as many triangle lists as possible into openGL display lists */
void _glassObjectMakeDisplayLists(GlassObject *object)
{
  int i;

  if(!object)
    return;

  for(i = 0; i < object->nlists; i++)
    _glassTriangleListMakeDisplayLists(object->list_array[i]);
}

/* Free an object */
void _glassFreeObject(GlassObject *object)
{
  if(!object)
    return;

  free(object->name);

  free(object->list_array);
  free(object->variable_array);
  free(object->apoint_array);

  free(object);
}
