/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>
#include <string.h>

#include "glass.h"

/* Make a texture */
GlassTexture *_glassTextureNew(void)
{
  GlassTexture *texture = malloc(sizeof(GlassTexture));

  if(texture == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for texture\n");
    return NULL;
  }

  texture->name = NULL;
  texture->is_alpha = FALSE;

  texture->image_file = NULL;
  texture->width = 0;
  texture->height = 0;

  texture->texture = 0;

  texture->next = NULL;

  return texture;
}

/* Set the texture's name */
void _glassTextureSetName(GlassTexture *texture, const char *name)
{
  free(texture->name);
  if(name == NULL)
    texture->name = strdup("");
  else
    texture->name = strdup(name);
}

/* Get the texture's name */
const char *_glassTextureGetName(GlassTexture *texture)
{
  return texture->name;
}

/* Load a RGB file, and use this as the texture */
void _glassTextureLoadRGB(GlassTexture *texture, const char *fname)
{
  GlassRGBImage *texture_image = _glassLoadRGBImage(fname);
  GLubyte *texture_data;
  GLenum data_type;
  int x, y, pad = 0;

  if(texture_image == NULL)
  {
    fprintf(stderr, "Error loading rgb texture :%s\n", fname);
    return;
  }

  free(texture->image_file);
  texture->image_file = strdup(fname);

  texture->width = texture_image->sizeX;
  texture->height = texture_image->sizeY;

  /* Allocate space for the texture data */
  /* !!note should really be able to be accessed later */
  texture_data = malloc(sizeof(GLubyte)
			* texture_image->sizeX
			* texture_image->sizeY
			* texture_image->sizeZ);

  /* Copy the data (from the image) to a new array */
  for(x = 0; x < texture_image->sizeX; x++)
  {
    for(y = 0; y < texture_image->sizeY; y++)
    {
      texture_data[(x + y * texture_image->sizeX) * texture_image->sizeZ]
	= texture_image->data[(x + texture_image->sizeX * y)
			      * texture_image->sizeZ + pad];
      texture_data[(x + y * texture_image->sizeX) * texture_image->sizeZ + 1]
	= texture_image->data[(x + texture_image->sizeX * y) *
			      texture_image->sizeZ + 1 + pad];
      texture_data[(x + y * texture_image->sizeX) * texture_image->sizeZ + 2]
	= texture_image->data[(x + texture_image->sizeX * y) *
			      texture_image->sizeZ + 2 + pad];
      if(texture_image->sizeZ > 3)
	texture_data[(x + y * texture_image->sizeX) * texture_image->sizeZ + 3]
	  = texture_image->data[(x + texture_image->sizeX * y) *
				texture_image->sizeZ + 3 + pad];
    }
    pad += (texture_image->sizeX * texture_image->sizeZ) % 4;
  }

  /* Generate the texture */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glDeleteTextures(1, &texture->texture);
  glGenTextures(1, &texture->texture);
  glBindTexture(GL_TEXTURE_2D, texture->texture);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  if(texture_image->sizeZ > 3)
  {
    data_type = GL_RGBA;
    texture->is_alpha = TRUE;
  }
  else
  {
    data_type = GL_RGB;
    texture->is_alpha = FALSE;
  }
  glTexImage2D(GL_TEXTURE_2D,
	       0,
	       texture_image->sizeZ,
	       texture_image->sizeX,
	       texture_image->sizeY,
	       0,
	       data_type,
	       GL_UNSIGNED_BYTE,
	       texture_data);

  /* The data is copied so remove the local version */
  free(texture_data);

  /* Free the image */
  free(texture_image->data);
  free(texture_image);
}

/* Free a glass texture, and remove the openGL texture it used */
void _glassFreeTexture(GlassTexture *texture)
{
  glDeleteTextures(1, &texture->texture);
  free(texture);
}
